#
# <20220518.1334.22>
#
# This routine presents the password locker passwords and options.
# It also allows editing of the entry.
#
proc editPasswords {} {
  # foreach par {
  #   ::pwLocker::pwAddress
  #   ::pwLocker::encryptRoutine
  #   ::pwLocker::decryptRoutine} {
  #   if {[set $par] == {}} {
  #     PopError "Password locker was either not or incorrectly initialized.\
  #              \nNothing here."
  #     return 
  #   }
  # }
  if {[info exists ::pwLocker::window] &&\
	  [winfo exists [set w $::pwLocker::window]] &&\
	  [$w.text edit modified] &&\
	  [yesNoCancel $w Really "An modified 'Edit Password' window exists.\
                                 \n Do you want to save it?"] == 0} {
    savePwChanges $w
  }

  if {[info exists ::pwLocker::window]} {
    destroy $w
  }
  # Lets use 'viewString' to show this...
  set whatToDo [list {Add a new Password} {addNewPw %w {} {} {}}\
		    {edit selected}   {editSelectedPass %w}\
		    {delete selected} {deleteSelectedPass %w}]
  set lines {}
  set w [ViewString "Passwords" lines boolean 0 utf16 $whatToDo spellOps 0]
  # Set window up to select by line
  bind $w <1> {set tk::Priv(selectMode) line
    catch {tk::TextSelectTo %W %x %y}
    catch {%W mark set insert sel.first}
  }
  set ::pwLocker::window $w
  populatePwEditWindow $w
}

proc populatePwEditWindow {w} {
  global glob config
  append lines  "Mouse <3> gives a menu which allows adding, editing "
  append lines  "or deleting a password. Mouse <1> selects lines. "
  append lines  "Entries may also be deleted with the 'Delete' key "
  append lines  "Mouse <3> deletes are immediately but 'Delete' key "
  append lines  "lines may be recalled with the 'Undo' and are saved "
  append lines  "only after a subsequent Mouse <3> delete or a Mouse <3> "
  append lines  "'saveChanges'.\n"
  append lines  "\n"

  append lines {    } ID\t
  append lines PASSWORD\t FLAGS\n
  set fon $glob(gui,ListBoxFont)
  set sz1 [font measure $fon "    ID"]
  set sz2 [font measure $fon "PASSWORD "]
  foreach ent [::pwLocker::patternGet *] {
    set arg [list {*}[lassign $ent id pw]]
    # set pw [eval [list {*}$::pwLocker::decryptRoutine $password]]
    set failsafe 0
    while {[lindex $pw end] == {} && $pw != {} && [incr failsafe] < 4} {
      set pw [lrange $pw 0 end-1]
    }
    append lines  $id\t [list {*}$pw]
    # We assume 'arg' the tail of the passed back pwLocker entry is a list
    # here we remove any empty strings
    
    if {[llength $arg] == 1} {
      set arg [regsub -all {^\{|\}$} $arg {}]
    }
    
    if {$arg != {}} {
      # If one of the entries is a drive letter ([a-z]:) we make it first
      # otherwise we don't touch it...
      if {[set la [lsearch -glob -nocase {[a-z]:} $arg]] > 0} {
	set arg [linsert $arg 0 [lindex $arg $la]]
	set arg [lrepace $arg $la+1 $la+1]
      }
      set narg ""
      frputs ln [llength $arg]
      foreach ar $arg {
	frputs arg ar narg ent
	if {$ar in {{ } {}} } continue
	lappend narg $ar
      }
      frputs narg arg
      if {$narg != ""} {
	incr flgPresent
	append lines \t $narg
      }
    }
    append lines \n
    lappend sz1 [font measure $fon "$id "]
    lappend sz2 [font measure $fon "[list {*}$pw] "]
    #frputs lines
  }
    frputs sz1 sz2
  set stp1 [eval "expr {max([join $sz1 {,}])}"]
  set stp2 [eval "expr {$stp1 + max([join $sz2 {,}])}"]
  $w.text delete 0.0 end
  fillTextWindow $w.text lines
  $w.text mark set insert 0.0
  $w.text edit reset
  $w.text edit modified 0
  ::autoscroll::autoscroll $w.scroll
  intelWinSize $config(geometry,textviewer) $w.text
  $w.text config -tabs "$stp1 $stp2"
  frputs stp1 stp2
  realWaitForIdle
  if {[info exists flgPresent]} {
    incr stp2 [font measure $fon "   hot-only   "]
  }
  lassign [split [wm geo $w] x+] wi h
  realWaitForIdle
  frputs stp1 stp2 w wi h
  if {$wi > $stp2} {
    wm geo $w ${stp2}x$h
  }
}

proc addNewPw {w id drive flag args} {
  # We ask the getpw routine to provide the password and unless we are way
  # wrong, it will pop up a dialog to fill in the details...
  # We are going to refresh the window after the return. If there are changes
  # let the user save those first.
  if {[$w.text edit modified]} {
    set f [smart_dialog .pwedit $w.text [_ "Save first?"]\
	       [list "Changes appear to have been made. Save these changes first?"]\
	       0 3 [list [_ "Yes"] [_ "No"] [_ "Cancel the add new request"]]]
    switch $f {
      0 {savePwChanges $w}
      2 {return 0}
      default {}
    }
  }
  
  set legal [list "no" "hot-only" "both"]
  set prompt [list "\nA full id consists of \"protocol://user@address/path\"\
             \nwhere \"://\" \"@\" and \"/\" are separators which appear as needed\
             \nThe \"user\" appears in the id after the protocol (and the \"://\"\
             \nseparator) and followed by e.g. \"@\"\
             \nIf \"address\" is nil \"localhost\" is used.\n\
             \nThe full password has three parts,\
             \n\"password\" \"private key disk address\" and \"pass phrase\"\
             \nany or all of these may be nil (\"\").\n\
	     \nFor MSW cifs shares the \"Drive\"  and \"Flag\" controls how the\
	     \nshare is connected \"Drive\" must be 'letter:' or '*:' or not given\
             \nWhere 'letter:' maps the share to that drive letter,\
             \nno drive letter sets the share up as a 'UNC' (//SYSTEM/SHARE).\
             \n\"Display Names\" may be generated for connected shares.\
             \nA flag to control the connection and Display Name has the following meaning.
	     \nNo    do not generate a Display Name.\
	     \nOnly  generate a Display Name (It will appear in the \"volume\" list).\
	     \nHot   generate a Display Name and put the display name in the HotList.\
	     \nDon't do not ever set up this share"]
	      
  set newpass [::pwLocker::getPw $id \
		   newuser 1 \
		   full 1 \
		   returnall 1 \
		   prompt $prompt \
		   {*}$args\
		   display [list {*}[expr {$::MSW ? [list "Drive" $drive] : {}}]\
				"Flag"  $flag]]
  if {$newpass in {0 -1 -2}} {return 0}
  array set newpw $newpass
  frputs "[array get newpw] " newpass
  set li ""
  # The drive and mapping info is for cifs passwords only
  if {[IsVFS $newpw(id)] && $VFSpro == "cifs"} {
    if {$newpw(Flag) != {}} {
      if {$::MSW &&\
	      [string index $newpw(Drive) 1] != ":" &&\
	      $newpw(Drive) != ""} {
	PopError "Expected \"Drive letter with colon or no drive\" but got\
               \"$newpw(Drive)\"\n(Flag requires a drive letter or nothing)"
	return 0
      }
      set li [lsearch -nocase -inline $legal $newpw(Flag)*]
      if {$li == {}} {
	PopError "Flag must be one of: \"$legal\" but found \"$newpw(Flag)\""
	return 0
      }
    }
    if {$::MSW} {
      if {$li != {}} {
	set li [list $newpw(Drive) $li]
      } else {
	set li $newpw(Drive)
      }
    }
  }
  ::pwLocker::putPw $newpw(id)\
      [list $newpw(password)\
	   $newpw(disk) \
	   $newpw(passphrase)]\
		  $li
  
  if {$args == {}} {
    populatePwEditWindow $w
  }
  return 1
}

proc deleteSelectedPass {w} {
  if {[catch {$w.text get sel.first sel.last} lines] != 0} {
    PopError "Problem reading window contents: $lines"
    return
  }
  frputs lines
  foreach line [split $lines \n] {
    # first toss the title line...
    if {$line == {} || [string index $line 0] == " "} {continue}
    lassign [split $line \t] id pw drive flag
    if {[set entries [::pwLocker::delete $id case 1]] != 1} {
      # Something bad going on. We should always get 1
      PopError "Delete of ID \"$id\" deleted $entries, should be exactly 1."
      break
    }
  }
  populatePwEditWindow $w
}


proc editSelectedPass {w} {
  if {[catch {$w.text get sel.first sel.last} lines] != 0} {
    PopError "Problem reading window contents: $lines"
    return
  }
  frputs lines
  foreach line [split $lines \n] {
    # first toss the title line...
    if {$line == {} || [string index $line 0] == " "} {continue}
    lassign [split $line \t] id pw drive flag
    set id [string trim $id]
    set pw [string trim $pw]
    lassign  $drive drive flag
    set flag [string trim $flag]
    if {!$::MSW} {
      set flag $drive
    }
    incr doneCount
    
    if {[addNewPw $w $id $drive $flag startwith $pw]} {continue}
    # Some sort of problem. If this is the first call, do nothing
    # otherwise repopulate the window to show what has been done
    if {$doneCount == 1} {return}
    break
  }
  populatePwEditWindow $w
 
}
# proc savePwChanges {w} {
#   # we will call pwLocker to save the passwords
#   # to preserve order we will back through the list as
#   # pwLocker always puts new entries first.
#   if {![$w.text edit modified]} {
#     PopInfo "Nothing to save."
#     return
#   }
#   if {[catch {$w.text get 3.0 end} lines] != 0} {
#     PopError "Problem reading window contents: $lines"
#     return
#   }
#   set $::pwLocker::pwAddress {}
#   set savefun $::pwLocker::saveFun
#   set ::pwLocker::saveFun {}
#   set ents [split $lines \n]
  
#   foreach line [lreverse $ents] {
#     if {$line == {} || [string index $line 0] == " "} {continue}
#     lassign [split $line \t] id pw flag
#     set id [string trim $id]
#     set pw [string trim $pw]
#     set flag [string trim $flag]
#     if {$id == "Mouse"} {break}
#     if {$id == "ID" && $pw == "PASSWORD" && $flag == "FLAGS"} {break}
#     ::pwLocker::putPw $id $pw $flag
#   }

#   set ::pwLocker::saveFun $savefun
#   eval [list {*}$savefun]
#   populatePwEditWindow $w

# }
