#
#<20230202.1329.45>
#
# This module contains bits of filerunner that are pure windows
# Putting this stuff here lightens the load on other platforms
#
# This bit will be sourced in global space

# This function (by name) triggers sourceing by the pkgIndex for
# the platform
startProc "First light in Win Platform"

if {[set it [info script]] ni $::auto_path} {
  lappend ::auto_path $it
}
unset -nocomplain it

proc load_now_fr_bits {} {}

# This is the turn on setup unique to this platform

set glob(os) WIN32
set ::MSW 1
set ::CASEops "-nocase"
# set ::TMP $::env(TMP)
# We want a unique name for the title, for speed and efficiency we will
# just encode the pid.
#
proc pidBase62 {} {
  set d ""
  set wpid [pid]
  while {$wpid > 0} {
    set rm [expr {$wpid % 62}]
    set wpid [expr {$wpid / 62}]
    append d [format %c [expr {$rm < 10 ? ($rm + 48) : $rm < 36 ? ($rm + 55) : ($rm + 61)}]]
  }
  return [string reverse $d]
}
set glob(orgTitle) "(fr #[pidBase62])"

proc buildTbarIcon {} {
  global glob
  wm iconbitmap . -default $glob(lib_fr)/icons/filerunner-icons.ico
  # This is called after init config so we have an opportunity do
  # platform specific config initialization.
  # We may want to rethink this later....
  # For now we will put these at the head of the button list
  
  # Here we add platform dependent buttons for the middle list.
  # format is the same as in fr.
  set properties {
    {{Properties} CmdProperties "" 0 \
	 {[_b "Brings up the MSW file properties window for selected file(s)." ]}}
    {{Short Cut} CmdCreateShortCut "" 0\
	 {[_b "Create a short cut for selected file(s)."]}}
    {{Short CutAS} CmdCreateShortCutAs "" 0\
	 {[_b "Create a short cut for selected file(s) with a new name."]}}
  }
  foreach platformButton $properties {
    lappend add [list [lindex $platformButton 0]]
    if {[lsearch -exact -index 0 $glob(cmds,list) [lindex $platformButton 0]] == -1} {
      lappend glob(cmds,list) $platformButton
    }
  }
  set ::config(middle_button_list) [linsert $::config(middle_button_list) 0 {*}$add]
}


startProc "After icon set up"

# if this is a wrapped process we move the twapi package to our
# private location to, well basically uncompress it.
if {[namespace exists freewrap]} {
  # If it exists and is the right version, leave well enough alone
  # Note that version is in the pkgIndex so we just compare the whole thing
  # We also skip this if twapi is built into the wrapper
  if {![file exists /zvfs/twapi]} {
    # not in the wrapper
    set twapiAdd [lindex [zvfs::list */twapi] 0]
    set r [catch {read [set fid [open $twapiAdd r]]} dat]
    set r2 [catch {read [set fid2 [open $glob(conf_dir)/twapi/pkgIndex.tcl r]]} dat2]
    catch {close $fid}
    catch {close $fid2}
    if {$r + $r2 != 0 || $dat != $dat2} {
      file copy -force [zvfs::list */twapi] $glob(conf_dir)/twapi
    } else {
      # puts "twapi ok"
    }
    lappend ::auto_path $glob(conf_dir)/twapi
  }
}

# puts "$::auto_path"
package require twapi
set ::Elevated [expr {[twapi::get_process_elevation] == "full"}]

startProc "After Twapi set up"

# We are agressive about cleaning up our tmp dir
proc cleanTmpFiles {} {
  global glob
  package require dde
  dde servername [pid]
  set servers [dde services TCLEval {}]
  foreach pid $servers {
    lappend pids [lindex $pid 1]
  }
  set dmess {}
  set glob(tmpdir) $glob(conf_dir)/tmp[pid]
  if {![file isdirectory $glob(tmpdir)]} {
    if {![file exists $glob(tmpdir)]} {
      set r [Try { file mkdir $glob(tmpdir) } "" 1]
      if {$r != 0} {
	lappend dmess [_ "Can't create %s. May be problems later" $glob(tmpdir)]
      }
    }
  }
  foreach dir [glob -tails -directory $glob(conf_dir) tmp*] {
    set who [regsub {tmp} $dir {}]
    if {$who == {} && [glob -nocomplain $glob(conf_dir)/tmp/*] != {}} {continue}
    if {$who ni $pids} {
      if {[catch {file delete -force $glob(conf_dir)/tmp$who} ms]} {
	lappend dmess $ms
      }
    }
  }
  if {$dmess != ""} {
    lappend ::mess "Error managing tmp files:"
    lappend ::mess {*}$dmess
  }
}
########################## end of general platform set up ##################
proc inotifySetUp {} {

}
# stubs for windows, used only in unix

proc GetUidGidString { file uid gid } {
    return ""
}

proc GetStringFromMode { mode } {
  global glob
  if { $::tcl_platform(platform) == "windows"} {
    return [format "0x%x" $mode]
  }
}

# The window kill function...
proc killInit {} {
  if {[catch {package require Tclx}] != 0} {
    proc kill {pids} {
      foreach pid $pids {
	twapi::end_process $pid -force
      }
      return {}
    }
  }
  return {}
}


# This routine is to get the full disc size. On
# *nix system it also returns the mount point of the
# containing file system


# Ok, take 2. Lets use the twapi com code to do this.
# our private global

# take 3.

# namespace eval ::oo {
#   variable fso  {}
#   variable drives {}
# }

proc GetDF { dir } {
  # frputs dir
  set ans [lindex [::twapi::get_volume_info [lindex [file split $dir] 0] -freespace]]
  #frputs ans
  if {[string index $ans end-1] == "."} {
    set ans [string range $ans 0 end-2]
  }
  #frputs ans
  return [list [formatDF $ans] {}]  
}


# This code in some senses emulates the linux version of inotify to
# handle the same functionallity on MSW.

# The watch takes:
# what   add or remove to set up or remove a watch
# which  the directory to monitor
# how    the watch flags

proc notifywatch {what which {how {}}} {
  global ::inotify::dirs
  # We only allow two monitors at a time, attempting more is
  # An error..
  if {$what == "add"} {
    if {[llength $dirs] >=2} {
      # frputs dirs which how
      return -code error "Attempting to monitor more than 2 dirs"
    }
    # frputs which how
    if {[catch {::twapi::begin_filesystem_monitor \
		    $which [list notecallback $which] {*}$how} out] != 0} {
      # frputs out which how
      return -code error $out
    }
    # sweet success...
    lappend dirs [list $which $out]
    return $which
  }
  # must be a remove
  set ndirs {}
  foreach dir $dirs {
    lassign $dir di han
    # frputs dir di han
    if {$di == $which} {
      ::twapi::cancel_filesystem_monitor $han
      incr found
      continue
    }
    lappend ndirs $dir
  }
  if {![info exists found]} {
    frputs  which dirs ndirs
    return -code error "$which not found in monitored directories"
  }
  set dirs $ndirs
}

proc notecallback {args} {
  global glob config
  set did 0
  set found {}
  # frputs args
  
  # For windows notecallback gets:
  # arg0   the directory we are monitoring
  # arg1   the monitor handle (not of any use here)
  # arg3   a list of dublets (where fd is the file name):
  # 'added' fd, 'removed' fd, 'modified' fd, 'renameold' fd, 'renamenew' fd
  # or 'error' error-code.
  # Of these we only seperate modified and error...
  lassign $args dir han dublets
  foreach lr {left right} {
    if {$glob(notify_id,$lr) == $dir} {
      foreach {type file} $dublets {
	set key [string index $type 0]
	if {$key == "m"} {
	  lappend glob(n_files,$lr) $file
	} elseif {$key != "e"} {
	  lappend found $lr
	  if {$key == "a" && [capture $ou] == 1 } {
	    proc capture { note } { }
	  }
	} else {
	  # We have an error, shut down the whole thing
	  notifywatch remove $dir
	  return -code error "Directory modify monitor error $file"
	}
      }
    }
  }
  doInotifyUpdate $found
}
############################### End of notify code ################
############################### build Tree stub    ################
proc buildTree {args} {}
############################### End of Tree stub   ################

############################### displays setup     ################
# Here we gather info on attached displays and hidden space in the
# windows (borders and titles).
#
namespace eval ::displays {
  # monitors is set at config time... don't change it
  variable monitors
}
package require registry
proc ::displays::init {} {
  # we use the registry to get the border info.  As near as we can tell the
  # title bar is composed of a "Caption" and a border. The "CaptionHeight",
  # "BorderWidth" and "PaddedBorderWidth" are all in the below referenced
  # registry key. The values are in "twips" which is a physical length and so
  # depends on DPI (aka "AppliedDPI") for its translation.
  # A twip is 1/1440 of an inch.
  # The height we calculate below is the Caption height + the PaddedBorderWith
  # (top and bottom of the title bar) + BorderWidth (bottom of window border)
  # We negate the constant since MSW keeps these as negative quanities.
  
  set key [file native {HKEY_CURRENT_USER/Control Panel/Desktop/WindowMetrics}]
  lassign {0 0} ::displays::monitors(height) ::displays::monitors(width)
  # First get the borders
  # If there is a problem with the registry (like in wine) we default to 0
  catch {
    foreach item {AppliedDPI BorderWidth PaddedBorderWidth CaptionHeight} {
      set $item [registry get $key $item]
    }
    set ::displays::monitors(height) [expr {($CaptionHeight + \
						 $BorderWidth +\
						 2 * $PaddedBorderWidth) *\
						$AppliedDPI / -1440}]
    set ::displays::monitors(width)  [expr {2 * $BorderWidth * $AppliedDPI / -1440}]
  }
  set ::displays::monitors(monitors) {}
  foreach display [twapi::get_multiple_display_monitor_info] {
    array set tmp $display
    lassign $tmp(-workarea) xmin ymin xmax ymax
    lappend ::displays::monitors(monitors) [list $xmin $xmax $ymin $ymax]
    array unset tmp
  }
  Log "Monitors found: [array get ::displays::monitors]"
  return 
}
# For MSW we do this now. It is fast and this means we don't have to save it
# in the config file
::displays::init
############################### end of displays setup ################


# fileNameTranslate returns a list of doublets in a format suitable for 
# 'array set'..  The following will only be present if needed:
#
# scheme {file nfs cifs}
# path   for file -$sfile, for nfs target file, for cifs part of $sfile to add
#                                                        to share location
# host   for file not needed, for nfs|cifs host ip 
# user   for file not needed, for nfs|cifs user used in fstab
#
# share  only for nfs & cifs the share name (host:/??) or (//host/???)
# If a vaue is unknown it will be {}.
proc fileNameTranslate {sfile} {
  set localLetter [string range $sfile 0 1]
  # This will throw an error for a local drive
  if {[catch {twapi::get_client_share_info\
		  $localLetter\
		  -remoteshare\
		  -user} shareData] == 0} {
    set scheme "cifs"
    array set tmp $shareData
    set share [regsub -all {\\} $tmp(-remoteshare) {/}]
    return [list path [string range $sfile 1 end]\
		share  $share\
		host $hostname \
		user tmp(-user)\
		scheme cifs\
		pwd cifs://$tmp(-user)@[string range $share 2 end]]
  } else {
    return [list path $sfile scheme file]
  }
}


proc setWinVolLetter {} {
}

proc getRemoteShares {} {
  set mes "This may take a while. There are two parts:
a) Find shares on the local network and
b) Analyze and compare with the connected shares and the password locker.

Each part will generate a progress bar at the top of filerunner's window."

  after 0  [list PopWarn $mes]
  after 1 ::UtilsMenu::enumShares

}
# Here we build an interactive window which displays each 'Share' we find on
# the network and all we know about it. We then allow interactions to change
# those things.

# Here we have a couple of entries for the Utilities menu
#
namespace eval ::UtilsMenu {
  # These entries are added to the utilities menu for Windows
  variable last {}
  # We want these in the order they appear in the table
  variable tablEntries [list server share drive pwDrive user password pwOk flags]
  variable readOnly {server share drive pwOk}
  variable perLine [llength $tablEntries] ; # Entries per line in our table
  variable colors 
  variable disAry
  variable disAryBackUp
  variable wayOut
  variable debug 0

  variable ents {
    {-label {Enumerate remote shares} -command {getRemoteShares}}
    {-label {Refresh PATH \n(after add/delete programs)} -command {::auto_execNS::buildPath}}
  }
  #  {-label {Assign drive letter} -command {setWinVolLetter}}
}
proc ::UtilsMenu::infoLineHead {add args} {
  variable perLine
  foreach arg $args {lappend result [list {} [list -text $arg {*}$add]]}
  return $result
}
proc ::UtilsMenu::infoLine {add text} {
  return [list [list fullspan [list -text $text {*}$add]]]
}

proc ::UtilsMenu::dataLine {entryNo args} {
  variable disAry
  variable readOnly
  variable colors
  
  foreach arg $args {
    lassign $arg ar option
    lappend  result\
	[list {}\
	     [list {*}[expr {$option != {} ? $option : {}}]\
		  -textvariable ::UtilsMenu::disAry($entryNo,$ar)\
		  {*}[expr {$ar in $readOnly ? [list -foreground  $colors(hlight)] : {}}]\
		  {*}[expr {$ar == "password" ? [list -show "*"] : {}}]]
	    ]
  }
  # frputs result
  return $result
}

# available drives list
proc ::UtilsMenu::availDr {availDr} {
  set taken [list {*}[file volume] {:/}]
  foreach dr {A B C D E F G H I J K L M N O P Q R S T U V W X Y Z {}} {
    if {"$dr:/" ni $taken} {
      lappend availDr $dr:
    }
  }
  return [lsort -unique $availDr]
}
# Test if the password exists and works
# parms password array to look through
# returns the 'user password drive Yes/No flags' as a list Yes password is good

proc ::UtilsMenu::findAndCheckPw {serv} {
  upvar pwInfoAry pwInfoAry
  set n -1
  set found 0
  lassign [split $serv /] > > server share
  set serv //[string toupper $server]/$share
  while {[info exists pwInfoAry($serv,[incr n],pw)]} {
    if {![info exists pwInfoAry($serv,$n,user)]} {continue}
    incr found
    break
  }
  if {!$found} {return {}}
  # See if it is a good user/password
  set testDr [lindex [availDr {}] 0]
  set r [catch {::twapi::connect_share [file native $serv]\
		    -user     $pwInfoAry($serv,$n,user)\
		    -password [twapi::conceal $pwInfoAry($serv,$n,pw)] \
		    -localdevice $testDr} error]
  frputs testDr n r serv error pwInfoAry($serv,$n,user)
  catch {twapi::disconnect_share $error -force -updateprofile}
  # If a not found error, assume the target is offline
  
  set rtn [list $pwInfoAry($serv,$n,user) $pwInfoAry($serv,$n,pw)\
	       $pwInfoAry($serv,$n,drive) \
	       [expr {$r == 0 ? "Yes" : \
			  [string match "*network*" $error] ? "Offline" : "No"}]\
	       $pwInfoAry($serv,$n,flags)]
  # Use this entry only once...
  unset pwInfoAry($serv,$n,user)
  return $rtn    
}

proc ::UtilsMenu::enumShares {} {
  variable disAry
  variable disAryBackUp
  variable pwInfoAry
  variable perLine
  variable debug
  variable wayOut
  variable last
  variable feedback {}  ; # text messages go here
  #                      status values nil   -> connected good password
  #                                    BadPw -> connected bad  password
  #                                    NoPw  -> connected no   password
  #                                    Found -> not connected ?? password
  #                                    Pw    -> not connected ?? password
  
  variable colors
  variable connectedShares
  variable wt1
  variable wt2
  variable w
  variable scrollWidth
  
  global foundShares
  global config
  array set colors [list \
			sbg      $config(gui,color_select_bg)\
			bg       $config(gui,color_bg)\
			hlight   $config(gui,color_highlight_fg)\
			hlightbg $config(gui,color_highlight_bg)\
			info     $config(gui,color_cmd)]
  
  if {![info exists ::foundShares]} {
    set foundSharesW {}
    foreach sh [::findUNC::UNCpoll] {
      lassign $sh serv dr
      lappend foundSharesW [list $dr $serv]
    }
    set ::foundShares [lsort $foundSharesW]
  }
  array unset pwInfoAry
  frputs foundShares
  
  while {1} {
    set foundSharesW $foundShares
    set connectedShares [lsort -index 1\
			     [regsub -all {\\} [twapi::wnet_connected_resources] {/}]]
    set last {}
    # WSL shares are always connected but, for unknown reasons do not appear here
    # unless they have a drive letter.

    foreach share $::findUNC::wslNativeMap {
      if {![string match -nocase *[lindex $share 0]* $connectedShares]} {
	lappend connectedShares [list {} [lindex $share 0]]
      }
    }
    set connectedShares [lsort $connectedShares]
    # Make a stab at the total count
    set shareNum [llength [lsort -unique [concat $foundShares $connectedShares]]] 
    frputs connectedShares
    # We have all shares that are on line at the moment.  There may be others
    # that we can not find due to the server being off line.  Some of those may
    # be in the password locker so lets get all of that info.
    
    set pws [pwLocker::patternGet cifs://*]
    # build an array of the bits we find
    frputs pws
    set pwList {}
    set pwsExp {}
    foreach pwEnt $pws {
      lassign $pwEnt id pw flags
      set splits [URLsplit $id]
      frputs splits 
      if {$splits == {}} {
	# Oops an entry that does not follow the rules. Save and report later
	lappen ruleBender $id
	continue
      }
      lassign {} drive flagsLst
      foreach item $flags {
	if {[string match -nocase {[a-z*]:} $item]} {
	  set drive [string toupper $item]
	} else {
	  lappend flagsLst $item
	}
      }
      lassign $splits pro user server port share
      set server [string toupper $server]
      lappend pwsExp [list $server $share $user $drive [lindex $pw 0] $flagsLst]
    }
    # Now we sort this list and arrange it in weighted order.
    # If a given server/share has more than one occurance we want the
    # most detailed entry first...
    set pwsExp [lsort $pwsExp]
    foreach pwEnt $pwsExp {
      lassign $pwEnt server share user drive pw flags
      set n -1
      # Handle dups duplicate, mark as dup
      while {[info exists pwInfoAry(//$server/$share,[incr n],pw)]} {}
      set pwInfoAry(//$server/$share,$n,user)  $user
      set pwInfoAry(//$server/$share,$n,pw)    $pw
      set pwInfoAry(//$server/$share,$n,flags) $flags
      set pwInfoAry(//$server/$share,$n,drive) $drive
      lappend pwList [list "//$server/$share"  $n]
      # Here we build the display array.
    }
    # Lets display what we have and open up changes to it.
    set columnHead [infoLineHead [list -bg $colors(info) -borderwidth 2 -relief sunken]\
			Server\n Share\n "Drive\nActual" "Drive\nRequest"\
			User\n Password\n "Password\n  OK" Flag\n]
    set head      [infoLine {} "Connected Shares with good password"]
    set headBadPw [infoLine {} "Connected Shares with bad password"]
    set headNoPw  [infoLine {} "Connected Shares with no password"]
    set headFound [infoLine {} "Unconnected Shares Found"]
    set headPw    [infoLine {} "Unused Share password entries"]
    set entryNo -1
    set rejects {}
    array unset disAry
    set connectedShares [lsort -index 1 $connectedShares] 
    frputs connectedShares
    lassign {} dataBadPw data dataNoPw dataFound dataPw
    LogStatusOnly {} 1

    foreach entry $connectedShares {
      incr entryNo
      
      LogStatusOnly "Working on $entry" [expr {($entryNo * 100) / $shareNum}]
      lassign $entry drv serv
      lassign [split $serv /] > > server share
      set server [string toupper $server]
      set drv [string toupper $drv]
      # Remove this share from the found-shares
      frputs foundSharesW server share drv serv
      set indx [lsearch -nocase -exact -index {1} $foundSharesW $serv]
      if {$indx != -1} {
	set foundSharesW [lreplace $foundSharesW $indx $indx]
	frputs foundSharesW
      }
      set disAry($entryNo,server) $server
      set disAry($entryNo,share)  $share
      set disAry($entryNo,drive)  $drv
      
      # now lets see what we know about this from the passwordLocker
      set pwInfo [findAndCheckPw $serv]
      lassign $pwInfo  disAry($entryNo,user) disAry($entryNo,password)\
	  disAry($entryNo,pwDrive) disAry($entryNo,pwOk)\
	  disAry($entryNo,flags)


       if {$pwInfo != {}} {	
	 set availDr [availDr [list $disAry($entryNo,pwDrive) *: {}]]
       } else {
	 set availDr [availDr [list *: {} $drv]]
       }
		     
      set entries\
	  [list server share drive [list pwDrive [list -values $availDr]]\
	       user password pwOk flags]
      set dataTmp [dataLine $entryNo {*}$entries]
      
      switch $disAry($entryNo,pwOk) {
	Yes {
	  set disAry($entryNo,stat) {}
	  set data [concat $data $dataTmp]
	}
	No  {
	  set disAry($entryNo,stat) BadPw
	  set dataBadPw [concat $dataBadPw $dataTmp]
	}
	default {
	  set disAry($entryNo,stat) NoPw
	  set dataNoPw [concat $dataNoPw $dataTmp]
	}
      }
    }
  
    
    # Now shares that are not connected...
    foreach share $foundSharesW {
      incr entryNo
      lassign $share drv serv
      LogStatusOnly {} 1
      LogStatusOnly "Working on $share" [expr {($entryNo * 100) / $shareNum}]
      lassign [split $serv /] > > server share
      set disAry($entryNo,server) [string toupper $server]
      set disAry($entryNo,share)  $share
      set disAry($entryNo,drive)  [string toupper $drv]
      # Look for a matching password
      set pwInfo [findAndCheckPw $serv]
      lassign $pwInfo  disAry($entryNo,user) disAry($entryNo,password)\
	  disAry($entryNo,pwDrive) disAry($entryNo,pwOk)\
	  disAry($entryNo,flags)


      set entries\
	  [list server share drive [list pwDrive [list -values [availDr {*: {}}]]]\
	       user password pwOk flags]
      set dataFound [concat $dataFound [dataLine $entryNo {*}$entries]]
      set disAry($entryNo,stat) Found
    }
    # Now do shares in password Locker
    set entries\
	[list server share drive [list pwDrive [list -values [availDr {*: {}}]]]\
	     user password pwOk flags]
    foreach pw $pwList {
      lassign $pw serv n
      if {[info exists pwInfoAry($serv,$n,user)]} {
	incr entryNo
	lassign  [findAndCheckPw $serv] disAry($entryNo,user) disAry($entryNo,password)\
	  disAry($entryNo,pwDrive) disAry($entryNo,pwOk)\
	  disAry($entryNo,flags)
	
	lassign [split $serv /] > > server share
	set disAry($entryNo,server)   $server
	set disAry($entryNo,share)    $share
	set disAry($entryNo,drive)    {}
	set dataPw [concat $dataPw [dataLine $entryNo {*}$entries]]
	lappend dataIndex [list $entryNo Pw]
	set disAry($entryNo,stat) Pw
      }
    }
    set exLines 6
    
    set groups {{} BadPw NoPw Found Pw}
    foreach post $groups {
      if {[set len$post [llength [set data$post]]] == 0} {
      set head$post {}
      incr exLines -1
      }
    }
    # Make an index of button numbers to entryNos.
    variable butToEnt {}
    
    set butStart [expr {($perLine + 1)}]
    
    # first button # of group
    set lastB {}
    set soFar 0
    set grpAry() $butStart
    foreach gp  [lrange $groups 1 end] {
      set grpAry($gp) [expr {$grpAry($lastB) + [set len$lastB] + [llength [set head$gp]]}]
      set lastB $gp
      set soFar$gp 0
    }
    set feedBack [infoLine [list -textvariable ::UtilsMenu::feedback\
				-bg white -fg red -bd 3 -relief sunken] {}]
    set entNo -1
    while {[info exists disAry([incr entNo],stat)]} {
      set grp $disAry($entNo,stat)
      lappend butToEnt [list [expr {([set soFar$grp] * $perLine) + $grpAry($grp)}] $entNo]
      incr soFar$grp
    }
    frputs "[array get grpAry] " butToEnt'
    
    # Lets build a scrollable window...
    set w .shares
    if {![winfo exists $w]} {
      toplevel $w
      set wt1 [text  $w.t1]
      set seGrip [segrip $w]
      grid $wt1 -in $w -column 0 -row 0 -rowspan 2 -sticky news
      scrollbar $w.scroll -relief sunken -command "$wt1 yview"
      grid $w.scroll -in $w -column 1 -row 0 -sticky ns
      grid $seGrip   -in $w -column 1 -row 1 -sticky sw
      $wt1 config -yscrollcommand "$w.scroll set"
      grid rowconfigure $w 0 -weight 1
      grid columnconfigure $w all  -weight 1
      grid columnconfigure $w 1    -weight 0
      set wt2 $wt1.t2
      bind $w <1> [list ::UtilsMenu::highlight %W]
      bind $wt1 <Configure> {::UtilsMenu::sizeAdjust %W}
      # The embedding seems to mess up the %W such that the
      # scroll commands get delivered to the embedded window
      # rather than the one that needs/uses them.
      bind $w <MouseWheel> [list ::UtilsMenu::scroll %D]
      bind $w <Shift-MouseWheel> [list ::UtilsMenu::scroll %D x]
      update
      set scrollWidth [winfo reqwidth $w.scroll]
      $w.segrip config -width $scrollWidth
    }
    wm deicon $w
    # set up the action buttons:
    set actButs [list \
		     [list "Changed as marked" [list -command ::UtilsMenu::changeMk]]\
		     [list "Connect share"     [list -command ::UtilsMenu::connect]]\
		     [list "Disconect share"   [list -command ::UtilsMenu::disconnect]]\
		     [list "Remove login/password" [list -command ::UtilsMenu::removePw]]\
		     [list "Show password"     [list -command ::UtilsMenu::showPw]]\
		     [list "Disconnect all shares\nand do turn on setup"\
			  [list -command ::UtilsMenu::disResart]]\
		     [list "Ok"]]

    # set bindings [list bind [list $wt2.0 <Button-3> [list frputs 0]]]
    set buttonCount [expr {($entryNo +2 ) * $perLine + [llength $actButs] + $exLines}]
    # Set up the readonly trace code...
    array set disAryBackUp [array get disAry]
    trace add variable ::UtilsMenu::disAry write ::UtilsMenu::readOnly 
    # frputs headCon' columnHead' buttonCount entryNo rejects data' data2'

    LogStatusOnly "done" -1
    
    set wayOut [smart_dialog $wt2 . {Shares}  [list text1 text2] -1 $buttonCount\
	       [list \
		    {*}$columnHead\
		    {*}$head\
		    {*}$data\
		    {*}$headBadPw\
		    {*}$dataBadPw\
		    {*}$headNoPw\
		    {*}$dataNoPw\
		    {*}$headFound\
		    {*}$dataFound\
		    {*}$headPw\
		    {*}$dataPw\
		    {*}$feedBack\
		    {*}$actButs]\
	       arrange {0 28 3} textfun ::UtilsMenu::textfun \
	       scroll 100 debug $debug ]
    
 
    $wt1 window create 0.0 -window $wt2
    update
    wm maxsize $w\
	[set reqw [expr {[winfo reqwidth $wt2] + $scrollWidth}]]\
	[winfo reqheight $wt2]

    cursor::propagate $wt2 arrow
    bind $wt2 <Configure> {::UtilsMenu::sizeAdjust %W}
    if {![info exist notFirstTime]} {
      incr notFirstTime
      lassign [split [wm geo .] x+] width height xloc yloc
      
      wm geo $w ${reqw}x[expr {$height - 150}]+${xloc}+[expr {$yloc + 150}]
    } else {
      wm geo $w $notFirstTime
      update
      $wt1 see end
    }
    frputs wayOut buttonCount
    vwait $wayOut
    frputs $wayOut
    if {[set $wayOut] == $buttonCount - 1 || [set $wayOut] == -1} {
      frputs $wayOut
      break
    }
    set notFirstTime [wm geo $w]
    wm withdraw $w
    destroy $wt2
    frputs  $wayOut
  }
  destroy $w
}

proc ::UtilsMenu::scroll {D {xy y}} {
  variable wt1
  
  if {$D >= 0} {
    $wt1 ${xy}view scroll [expr {-$D/3}] pixels
  } else {
    $wt1 ${xy}view scroll [expr {(2-$D)/3}] pixels
  }
}

proc ::UtilsMenu::sizeAdjust {args} {
  variable wt1 ; # The text window which contains
  variable wt2 ; # this window.
  variable w
  variable scrollWidth 
  
  # size changes come from two user actions
  # 1st from wt2 as a result of overt window size changes and
  # 2ed from entries (primarily in the Flags column)
  # we also get an inital call after wt2 is built
  if {[winfo exists $wt2]} {
    wm maxsize $w\
	[expr {[winfo reqwidth $wt2] + $scrollWidth}]\
	[winfo reqheight $wt2]
  }

}

proc ::UtilsMenu::readOnly {Ary elem dc} {
  variable disAry
  variable disAryBackUp
  variable readOnly
  set id [string range $elem [string first "," $elem]+1 end]
  if {$id in $readOnly} {
    set disAry($elem) $disAryBackUp($elem)
  }
  
}

# bind and action routines

proc ::UtilsMenu::highlight {what} {
  variable last
  variable curEnt
  variable butToEnt
  variable perLine
  variable colors
  variable feedback {}
  frputs what
  if { ![regexp {^(\..*\.)([0-9]+)$} $what -> w butNo]} {
    frputs  "off the map "
    return
  }
  set butNo [string range $what [string last . $what]+1 end]
  foreach cent $butToEnt {
    lassign $cent but ent
    if {!($butNo >= $but && $butNo < $but + $perLine)} {continue}
    # Ok the entry number is $ent
    # lets try and highlight the line
    for {set bNum $but} {$bNum < $but + $perLine} {incr bNum} {
      $w$bNum config -background $colors(sbg)
    }
    if {$last != {} && $last != $but} {
      frputs last but curEnt ent
     
      for {set bNum $last} {$bNum < $last + $perLine} {incr bNum} {
	$w$bNum config -background $colors(bg)
      }
      ::UtilsMenu::showPw 1
    }
    
    frputs what butNo ent last but
    set last $but
    set curEnt $ent
    return
  }
  frputs "Item ($butNo) is not selectable " 
}

proc ::UtilsMenu::showPw {{hide {}}} {
  variable last
  variable tablEntries 
  variable wt2
  variable feedback

  if {[info exists last] && $last != {}} {
    set butNo [expr {$last + [lsearch -exact $tablEntries "password"]}]
    if {[$wt2.$butNo cget -show] == {} || $hide != {}} {
      $wt2.$butNo config -show {*}
    } else {
      $wt2.$butNo config -show {}
    }
    return
  }
    set feedback "ShowPw: Nothing selected."
}


# change the old entry to the new one as marked
proc ::UtilsMenu::changeMk {} {
  variable last
  variable curEnt
  variable butToEnt
  variable perLine
  variable disAry
  variable disAryBackUp
  variable readOnly 
  variable tablEntries
  variable wayOut
  variable feedback {}

  # 'last' is a button number for the first entry in the highlighted line
  # turn that into the entry number
  if {[info exists last] && $last != {} && [info exists disAry($curEnt,stat)]} {
    foreach element $tablEntries {
      if {$disAry($curEnt,$element) == $disAryBackUp($curEnt,$element)} {continue}
      # Ok something changed. We need to build a new password entry.
      # if the user changed, we need to first remove the old one.
      if {$disAryBackUp($curEnt,user) == {}} {
	::pwLocker::delete\
	    cifs://$disAry($curEnt,server)/$disAry($curEnt,share)\
	    -nocase 1
      } else {
	::pwLocker::delete\
	    cifs://$disAryBackUp($curEnt,user)@$disAry($curEnt,server)/$disAry($curEnt,share)\
	    -nocase 1
      }
      # put new one in the locker
      set userAt [expr {$disAry($curEnt,user) == {} ?\
			    [list {}] : [list "$disAry($curEnt,user)@"]}]
      if {$disAry($curEnt,user) == {}} {
	::pwLocker::putPw\
	    cifs://$disAry($curEnt,server)/$disAry($curEnt,share)\
	    $disAry($curEnt,password)\
	    {*}[list {*}$disAry($curEnt,pwDrive) {*}$disAry($curEnt,flags)]
      } else {
	::pwLocker::putPw\
	    cifs://$disAry($curEnt,user)@$disAry($curEnt,server)/$disAry($curEnt,share)\
	    $disAry($curEnt,password)\
	    {*}[list {*}$disAry($curEnt,pwDrive) {*}$disAry($curEnt,flags)]
      }
      set $wayOut 100
      return
    }
    set feedback "Change As Marked: Nothing changed."
    return
    frputs "no change "
  }
  set feedback "Change As Marked: Nothing selected."
  frputs "nothing highlighed "
      # If here, non selected or ???
      # just return
}

# This function attempts to connect the highlighted share.
# If already connected it does nothing. It ignores any edits to the
# share.

proc ::UtilsMenu::connect {} {
  variable last
  variable curEnt
  variable disAry
  variable disAryBackUp
  variable tablEntries
  variable feedback {}
  variable wayOut
  # 'curEnt' is the target entry
  frputs curEnt last
  if {[info exists last] && $last != {} && [info exists disAry($curEnt,stat)]} {
    set stat $disAry($curEnt,stat)
    if {$stat in {Found Pw} ||\
	    ($disAry($curEnt,server) == {WSL$} &&\
		 $disAry($curEnt,drive) == {} &&\
		 $disAry($curEnt,pwDrive) != {})} {
    } else {
      set feedback "Connect: Already connected! Abandoning connect call!"
      frputs feedback
      return
    }
    foreach element $tablEntries {
      if {$disAry($curEnt,$element) == $disAryBackUp($curEnt,$element)} {continue}
      # something changed.
      set feedback "Entry changed. Abandoning connect call."
      frputs feedback
      return
    }
    # Ok no changes
    set r [catch {::twapi::connect_share\
		      [file native //$disAry($curEnt,server)/$disAry($curEnt,share)]\
		      -user     $disAry($curEnt,user)\
		      -password [twapi::conceal $disAry($curEnt,password)] \
		      -localdevice $disAry($curEnt,pwDrive)} error]
    if {$r != 0} {
      set feedback "Connect failed with ($error)."
    } else {
      set $wayOut 100
      return
    }

  }
  set feedback "Connect: Nothing selected."
}
proc ::UtilsMenu::disconnect {} {
  variable curEnt
  variable last
  variable disAry
  variable feedback {}
  variable wayOut
  if {[info exists last] && $last != {} && [info exists disAry($curEnt,stat)]} {
    set stat $disAry($curEnt,stat)
    if {$stat ni {{} BadPw NoPw}} {
      set feedback "Disconnect: Not connected! Abandoning disconnect call!"
      frputs stat "Already disconnected. Abandoning disconnect call. "
      return
    }
    if {$disAry($curEnt,server) == {WSL$} && $disAry($curEnt,drive) == {}} {
      set feedback "Disconnect: 'WSL$' shares without drive letter can not be disconnected."
      return
    }
    set id [expr {$disAry($curEnt,drive) != {} ? $disAry($curEnt,drive) :\
		      [file native //$disAry($curEnt,server)/$disAry($curEnt,share)]}]
    if {[catch {twapi::disconnect_share $id -force -updateprofile} error] != 0} {
      set feedback "Disconnect: Disconnect of '$id' failed with error '$error'"
    } else {
      set feedback "'$id' successfully disconnected."
      set $wayOut 100
    }
  }
  set feedback "Disconnect: Nothing selected."
}
# Here we disconnect all and do the startup thing
proc ::UtilsMenu::disResart {} {
  variable connectedShares
  variable wayOut
  
  foreach share $connectedShares {
    lassign $share drv serv
    set id [expr {$drv != {} ? $drv : [file native $serv]}]
    catch  {twapi::disconnect_share $id -force -updateprofile}
  }
  setUpShare
  set $wayOut 100
}

proc ::UtilsMenu::removePw {} {
  variable curEnt
  variable last
  variable disAry
  variable disAryBackUp
  variable tablEntries
  variable feedback {}
  variable wayOut
  
  if {[info exists last] && $last != {} && [info exists disAry($curEnt,stat)]} {
    foreach element $tablEntries {
      if {$disAry($curEnt,$element) == $disAryBackUp($curEnt,$element)} {continue}
      set feedback "RemovePw: '$element' changed. Please try again without changing anything."
      return
    }
    set target "cifs://"
    if {$disAry($curEnt,user) != ""} {
      append target "$disAry($curEnt,user)@"
    }
    append target $disAry($curEnt,server)/$disAry($curEnt,share)
    if {[::pwLocker::delete $target -nocase 1] != 1} {
      set feedback "Failed to remove $target"
      return
    } 
    set feedback "RemovePw: Sucess! '$target' removed."
    
    set $wayOut 100
    return    
  }
  set feedback "RemovePw: Nothing selected."
}

proc ::UtilsMenu::textfun {w} {
  variable colors
  frputs w
  $w insert insert "This window shows the status of the available shares at this time. However, if a server is offline, its share may not be here or its password entry may be the only part available.\n
You may use the action buttons to edit the elements of the 'password', i.e the 'Request Drive' 'User' 'Password' and 'Flags'. Only drive letters in the 'Request Drive' drop down are allowed. This may mean you will need to free a drive letter from another share before using it.

After each action is taken either the window is withdrawn and rebuilt or an error message will appear just above the action buttons.\n"
  catch {font delete tm}
  $w tag configure bold -font \
      [font create tm {*}[font actual [$w cget -font]] -weight bold]

  $w insert insert\
      "\nThe 'Disconnect share' action disconnects the share even if it is use!!\n" bold
  
  $w tag configure select -background $colors(sbg)
  $w tag configure bg -background $colors(bg)
  $w tag configure readonly -foreground $colors(hlight)
  $w tag configure head -background $colors(info)
  $w tag configure feedback -background $colors(hlightbg) -foreground $colors(hlight)
 
  $w insert insert "\nColor meanings in the below grid:\n"
  $w insert insert "This is the normal background color (Entry BG).\n" bg
  $w insert insert "This is the column head background color (Shell Cmd).\n" head
  $w insert insert "This is the read-only foreground color (Highlight FG).\n" readonly
  $w insert insert "This is the selected background color (Selection BG).\n" select
  $w insert insert "This is the feedback line coloring (Selection BG Highlight FG).\n" feedback
  $w insert insert "\nThe flag field indicates if filerunner should not generate a 'Display Name' for this share (the default is to generate a 'Display Name'). A share is a 'volume' so it will appear in the 'volume' directory either as its 'Drive' letter or its 'Display Name' (if generated). The flag field allows you to include it or not in the 'Hotlist'.

Flag meanings are:
 No    do not generate a 'Display Name'.\n\
 Only  generate a 'Display Name' (It will appear in the 'volume' list).\n\
 Hot   generate a 'Display Name' and put the 'Display Name' in the 'HotList'.\n\
 Don't do not ever set up this share.\n
  Note: The flag meanings are not enforced in this window.
  Note: Only the first letter of the 'Flag' entry is checked.
  Note: Flag values that start with other than {N O H D} are ignored 
        and may be used to keep notes as the user desires.

(\"Shift-Mmousewheel\" scrolls horizontally  \"Mouse-Wheel\" scrolls vertically) "
}

